/*******************************************************************************
 * Copyright (c) 2013-2015 Pablo Pavon-Marino, Jose-Luis Izquierdo-Zaragoza.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the GNU Lesser Public License v3
 * which accompanies this distribution, and is available at
 * http://www.gnu.org/licenses/lgpl.html
 *
 * Contributors:
 *     Pablo Pavon-Marino, Jose-Luis Izquierdo-Zaragoza - initial API and implementation
 ******************************************************************************/
package com.net2plan.examples.netDesignAlgorithm.fa;

import cern.colt.matrix.tdouble.DoubleMatrix2D;
import com.net2plan.interfaces.networkDesign.IAlgorithm;
import com.net2plan.interfaces.networkDesign.Net2PlanException;
import com.net2plan.interfaces.networkDesign.NetPlan;
import com.net2plan.libraries.GraphUtils;
import com.net2plan.libraries.IPUtils;
import com.net2plan.utils.CollectionUtils;
import com.net2plan.utils.Constants.OrderingType;
import com.net2plan.utils.Constants.SearchType;
import com.net2plan.utils.DoubleUtils;
import com.net2plan.utils.Pair;
import com.net2plan.utils.Triple;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.ListIterator;
import java.util.Map;
import java.util.Random;
import java.util.Set;

/**
 * This algorithm computes the OSPF routing that minimizes the average congestion 
 * between (i) the congestion when no failure occurs in the network, (ii) the worse 
 * congestion ranging the cases when one single SRG is failed. If no SRGs are defined, 
 * congestion in (i) is returned. An evolutionary algorithm is implemented. Each 
 * solution is coded as an array of as many elements as links, storing the OSPF 
 * weight in each link. Details on the particular form in which the evolutionary 
 * operators are implemented can be seen in the source code. The returning design 
 * includes the routes defined by the OSPF weights, and an attribute per link with 
 * its OSPF weight.
 * 
 * @author Pablo Pavon-Marino, Jose-Luis Izquierdo-Zaragoza
 * @version 1.1, May 2015
 */
public class FA_EA_minCongestionSingleFailureOSPF implements IAlgorithm
{
	@Override
	public String executeAlgorithm(NetPlan netPlan, Map<String, String> algorithmParameters, Map<String, String> net2planParameters)
	{
		/* Basic checks */
		int N = netPlan.getNumberOfNodes();
		int E = netPlan.getNumberOfLinks();
		int D = netPlan.getNumberOfDemands();
		if (N == 0 || D == 0 || E == 0) throw new Net2PlanException("This algorithm requires a topology with links, and a demand set");

		/* Initialize some variables */
		final int maxLinkWeight = Integer.parseInt(algorithmParameters.get("maxLinkWeight")); /* number of loopless candidate paths per demand */
		if (maxLinkWeight < 1) throw new Net2PlanException("The maximum link weight must be greater or equal than one" + maxLinkWeight);
		final double maxExecTimeSecs = Double.parseDouble(algorithmParameters.get("maxExecTimeSecs"));
		final int ea_populationSize = Integer.parseInt(algorithmParameters.get("ea_populationSize"));
		final int ea_offSpringSize = Integer.parseInt(algorithmParameters.get("ea_offSpringSize"));
		final double ea_fractionParentsChosenRandomly = Double.parseDouble(algorithmParameters.get("ea_fractionParentsChosenRandomly"));
		if (ea_offSpringSize > ea_populationSize) throw new Net2PlanException("The offspring size cannot exceed the population size divided by two");

		EvolutionaryAlgorithmCore gac = new EvolutionaryAlgorithmCore(netPlan, maxExecTimeSecs, maxLinkWeight, ea_populationSize, ea_fractionParentsChosenRandomly, ea_offSpringSize);
		gac.evolve();

		/* Save the best solution found into the netPlan object */
		Map<Long, Double> linkWeightMap = gac.best_solution;
		IPUtils.setLinkWeights(netPlan, linkWeightMap);
		IPUtils.setECMPForwardingRulesFromLinkWeights(netPlan, linkWeightMap);

		return "Ok! Cost: " + gac.best_cost;
	}

	@Override
	public String getDescription()
	{
		return "This algorithm computes the OSPF routing that minimizes the average congestion between (i) the congestion when no failure occurs in the network, (ii) " + "the worse congestion ranging the cases when one single SRG is failed. If no SRGs are defined, congestion in (i) is returned. An evolutionary algorithm is " + "implemented. Each solution is coded as an array of as many elements as links, storing the OSPF weight in each link. "
			+ "Details on the particular form in which the evolutionary operators are implemented can be seen in the source code. " + "The returning design includes the routes defined by the OSPF weights, and an attribute per link with its OSPF weight";
	}

	@Override
	public List<Triple<String, String, String>> getParameters()
	{
		List<Triple<String, String, String>> algorithmParameters = new LinkedList<Triple<String, String, String>>();
		algorithmParameters.add(Triple.of("maxLinkWeight", "10", "Maximum OSPF weight associated to a link"));
		algorithmParameters.add(Triple.of("maxExecTimeSecs", "10", "Execution time of the algorithm. The algorithm will stop after this time, returning the best solution found"));
		algorithmParameters.add(Triple.of("ea_populationSize", "1000", "Number of elements in the population"));
		algorithmParameters.add(Triple.of("ea_offSpringSize", "500", "Number of childs in the offspring every generation"));
		algorithmParameters.add(Triple.of("ea_fractionParentsChosenRandomly", "0.5", "Fraction of the parents that are selected randomly for creating the offspring"));

		return algorithmParameters;
	}

	private static class EvolutionaryAlgorithmCore
	{
		private final Set<Long> nodeIds, linkIds, srgIds;
		private final Map<Long, Integer> linkId2LinearIndex;
		private final int E;
		private final long[] linkIdsVector;
		private final Map<Long, Pair<Long, Long>> linkMap, demandMap;
		private final Map<Long, Double> u_e, h_d;
		private final Map<Long, Set<Long>> srgLinksMap;
		private final long algorithmEndTime;
		private final int maxLinkWeight;
		private final int ea_populationSize;
		private final double ea_fractionChosenRandomly;
		private final int ea_offspringSize;
		private final Random rnd;
		private List<Map<Long, Double>> population;
		private double[] costs;
		public double best_cost;
		public Map<Long, Double> best_solution;

		/**
		 * Initialize the object with the problem data and the genetic algorithm specific parameters.
		 * 
		 * @since 1.0
		 */
		private EvolutionaryAlgorithmCore(NetPlan netPlan, double maxExecTimeSecs, int maxLinkWeight, int ea_populationSize, double ea_fractionParentsChosenRandomly, int ea_offSpringSize)
		{
			nodeIds = netPlan.getNodeIds();
			linkIds = netPlan.getLinkIds();
			srgIds = netPlan.getSRGIds();
			linkIdsVector = netPlan.getLinkIdsVector();
			E = linkIdsVector.length;
			linkId2LinearIndex = CollectionUtils.convertId2LinearIndexMap(linkIds);
			linkMap = netPlan.getLinkMap();
			demandMap = netPlan.getDemandMap();
			u_e = netPlan.getLinkCapacityMap();
			h_d = netPlan.getDemandOfferedTrafficMap();
			srgLinksMap = computeSRGLinksMap(netPlan);

			this.maxLinkWeight = maxLinkWeight;
			this.ea_populationSize = ea_populationSize;
			this.ea_fractionChosenRandomly = ea_fractionParentsChosenRandomly;
			this.ea_offspringSize = ea_offSpringSize;
			rnd = new Random();
			algorithmEndTime = System.nanoTime() + (long) (maxExecTimeSecs * 1E9);
		}

		/**
		 * Auxiliary function: Computes the congestion in a network for a given 
		 * set of OSPF weights. Those links with a weight equal to Double.MAX_VALUE 
		 * are supposed to not exist => the routing behaves as if those links did 
		 * not exist.
		 *
		 * @since 1.0
		 */
		private double computeCongestion(Map<Long, Double> solution_e)
		{
			/* Compute the congestion when there are no failures */
			DoubleMatrix2D f_te = IPUtils.computeECMPRoutingTableMatrix(nodeIds, linkMap, solution_e);

			List<Long> demands_p = new LinkedList<Long>();
			List<Double> x_p = new LinkedList<Double>();
			List<List<Long>> seqLinks_p = new LinkedList<List<Long>>();
			GraphUtils.convert_fte2xp(nodeIds, linkMap, demandMap, h_d, f_te, demands_p, x_p, seqLinks_p);
			
			double[] y_e = GraphUtils.convert_xp2ye(linkMap, x_p, seqLinks_p);
			double congestion = 0;
			for(long linkId : linkIds)
			{
				int e = linkId2LinearIndex.get(linkId);
				double rho_e = y_e[e] == 0 ? 0 : y_e[e] / u_e.get(linkId);
				if (rho_e > congestion) congestion = rho_e;
			}
			
			return congestion;
		}
		
		/**
		 * Computes the cost of a given solution.
		 * 
		 * @since 1.0
		 */
		private double computeCostSolution(Map<Long, Double> solution_e)
		{
			final double congestion_noFailure = computeCongestion(solution_e);
			if (srgIds.isEmpty()) return congestion_noFailure;
			
			Map<Long, Double> congestionPerSRG = new LinkedHashMap<Long, Double>(); /* congestion when each SRG fails */
			for(long srgId : srgIds)
			{
				Set<Long> linkIds_thisSRG = srgLinksMap.get(srgId);
				Map<Long, Double> aux_solution_e = new LinkedHashMap<Long, Double>(solution_e);
				for(long linkId : linkIds_thisSRG) aux_solution_e.put(linkId, Double.MAX_VALUE);
				
				congestionPerSRG.put(srgId, computeCongestion(aux_solution_e));
			}
			
			return 0.5 * congestion_noFailure + 0.5 * DoubleUtils.maxValue(congestionPerSRG);
		}

		/**
		 * Auxiliary function: For each SRG, computes the list of links that are 
		 * down when the SRG fails. This includes the links belonging to the SRG, 
		 * and the incoming and outgoing links of the nodes belonging to the SRG.
		 *
		 * @since 1.0
		 */
		private Map<Long, Set<Long>> computeSRGLinksMap(NetPlan netPlan)
		{
			Map<Long, Set<Long>> out = new LinkedHashMap<Long, Set<Long>>();
			for(long srgId : netPlan.getSRGIds())
			{
				Set<Long> linkIds_thisSRG = new LinkedHashSet<Long>(netPlan.getSRGLinks(srgId));
				Set<Long> nodeIds_thisSRG = netPlan.getSRGNodes(srgId);
				for(long nodeId : nodeIds_thisSRG) linkIds_thisSRG.addAll(netPlan.getNodeAssociatedLinks(nodeId));

				out.put(srgId, linkIds_thisSRG);
			}
			
			return out;
		}

		/**
		 * This method is called after creating the object, to launch the genetic 
		 * algorithm. The method will first create an initial population. Then, 
		 * will complete a number of iterations given by ea_numberIterations. In each 
		 * iteration, a population is transformed using the evolutionary operators: 
		 * parent selection, crossover, mutation, selection. The method keeps track 
		 * of the best solution found, which is kept in the internal variables 
		 * best_solution, best_cost. Then, they are publicly accessible when the 
		 * method returns.
		 * 
		 * @since 1.0
		 */
		public void evolve()
		{
			/* Generate the initial population */
			generateInitialSolutions();

			/* Update the best solution found so far */
			final int bestSolutionId = DoubleUtils.maxIndexes(costs, SearchType.FIRST)[0];
			best_cost = costs[bestSolutionId];
			best_solution = new LinkedHashMap<Long, Double>(population.get(bestSolutionId));
			System.out.println("Initial population. Best solution cost: " + best_cost);

			/* Evolve: one iteration per generation */
			int it = 0;
			while (System.nanoTime() < algorithmEndTime)
			{
				LinkedList<Integer> parents = operator_parentSelection();
				List<Map<Long, Double>> offspring = operator_crossover(parents);
				this.operator_mutate(offspring);
				this.operator_select(offspring);
				if (costs[0] < best_cost)
				{
					best_cost = costs[0];
					best_solution = new LinkedHashMap<Long, Double>(this.population.get(0));
				}
				
				System.out.println("Iteration: " + it + ". Best solution cost: " + best_cost + ", best solution: " + best_solution);
				it++;
			}
		}

		/**
		 * Generates an initial population of solutions. Each solution is generated 
		 * choosing randomly for each link, a weight within the range 1...maxLinkWeight.
		 * 
		 * @since 1.0
		 */
		private void generateInitialSolutions()
		{
			population = new ArrayList<Map<Long, Double>>(ea_populationSize);
			costs = new double[ea_populationSize];

			for (int cont = 0; cont < ea_populationSize; cont++)
			{
				Map<Long, Double> solution_e = new LinkedHashMap<Long, Double>();
				for(long linkId : linkIds)
					solution_e.put(linkId, 1.0 + (double) rnd.nextInt(maxLinkWeight));

				population.add(solution_e);
				costs[cont] = computeCostSolution(solution_e);
			}
		}

		/**
		 * Receive a set of parents of size ea_offspringSize * 2. From them, 
		 * generates an offspring of size ea_offspringSize. Couples are generated 
		 * randomly, each parent appears in only one couple. Given two parents, 
		 * the descendant is created by, for each link, picking from any parent 
		 * (chosen randomly), the link weight.
		 * 
		 * @since 1.0
		 */
		private List<Map<Long, Double>> operator_crossover(LinkedList<Integer> parents)
		{
			/* The offspring to be returned */
			List<Map<Long, Double>> offspring = new ArrayList<Map<Long, Double>>(parents.size() / 2);

			/* Shuffle randomly the parent selection. Then, couples are formed randomly */
			Collections.shuffle(parents);

			/* Two consecutive parents are a couple */
			while (parents.size() >= 2)
			{
				final int firstParentId = parents.poll();
				final int secondParentId = parents.poll();

				final Map<Long, Double> firstParent = population.get(firstParentId);
				final Map<Long, Double> secondParent = population.get(secondParentId);

				/* The descendant chooses randomly for each link, the link weight from any of the two parents */
				final Map<Long, Double> descendant = new LinkedHashMap<Long, Double>();
				for (long linkId : linkIds) descendant.put(linkId, rnd.nextBoolean() ? firstParent.get(linkId) : secondParent.get(linkId));
				offspring.add(descendant);
			}
			
			return offspring;
		}

		/**
		 * Receives a population (typically an offspring). For each element in the 
		 * population, mutates the solution. A mutation consists of choosing a link 
		 * randomly, and changing randomly the link weight in the range 1...maxLinkWeight.
		 * 
		 * @since 1.0
		 */
		private void operator_mutate(List<Map<Long, Double>> offspring)
		{
			for (Map<Long, Double> solution : offspring)
			{
				final int e = rnd.nextInt(E);
				final long linkId = linkIdsVector[e];
				
				solution.put(linkId, 1.0 + rnd.nextInt(maxLinkWeight));
			}
		}

		/**
		 * Selects from the population ea_offspringSize*2 elements that will 
		 * become parents, generating the offspring. A fraction of the parents 
		 * (1-ea_fractionChosenRandomly) is chosen from the best in the population. 
		 * A fraction ea_fractionChosenRandomly is chosen randomly from all the 
		 * population, so that repetitions can occur.
		 *
		 * @since 1.0
		 */
		private LinkedList<Integer> operator_parentSelection()
		{
			final int numParentsToChoose = 2 * ea_offspringSize;
			final int numParentsChosenRandomly = (int) Math.floor(numParentsToChoose * ea_fractionChosenRandomly);
			final int numParentsChosenFromCost = numParentsToChoose - numParentsChosenRandomly;

			/* Initialize the list to be returned */
			LinkedList<Integer> chosenParents = new LinkedList<Integer>();

			/* Sorts the existing population: first the best cost */
			sortAscending(this.population, this.costs);

			/* Choose the best solutions as parents: as many as numParentsChosenFromCost */
			for (int cont = 0; cont < numParentsChosenFromCost; cont++)
			{
				chosenParents.add(cont);
			}

			/* The rest of the parents (numParentsChosenRandomly) are chosen randomly among the non-selected. Parents can be repeated */
			for (int cont = 0; cont < numParentsChosenRandomly; cont++)
			{
				chosenParents.add(rnd.nextInt(this.population.size()));
			}

			return chosenParents;
		}

		/**
		 * Given the population from previous generation, and a new offspring, 
		 * selects the elements that will pass to the next generation. These will 
		 * be the best ea_populationSize among all the aggregated population.
		 * 
		 * @since 1.0
		 */
		private void operator_select(List<Map<Long, Double>> offspring)
		{
			/* Compute the costs of the offspring */
			double[] offspringCosts = new double[offspring.size()];
			ListIterator<Map<Long, Double>> it = offspring.listIterator();
			while(it.hasNext()) offspringCosts[it.nextIndex()] = computeCostSolution(it.next());

			/* Append the offspring to the population */
			population.addAll(offspring);
			costs = DoubleUtils.concatenate(costs, offspringCosts);

			/* Reorder the new population */
			sortAscending(population, costs);

			/* Take the best ones, keeping the population size equal to ea_populationSize */
			while (population.size() > ea_populationSize)
			{
				population.remove(population.size() - 1);
			}
			
			costs = Arrays.copyOf(costs, ea_populationSize);
		}

		/**
		 * Auxiliary function: Sorts a population in ascending order according to 
		 * its cost (best cost first).
		 * 
		 * @since 1.0
		 */
		private void sortAscending(List<Map<Long, Double>> population, double[] costs)
		{
			double[] copyCosts = Arrays.copyOf(costs, costs.length);
			List<Map<Long, Double>> copyPopulation = new ArrayList<Map<Long, Double>>(population);

			int[] orderedIndexes = DoubleUtils.sortIndexes(costs, OrderingType.ASCENDING);
			for (int newPosition = 0; newPosition < orderedIndexes.length; newPosition++)
			{
				final int oldIndex = orderedIndexes[newPosition];
				costs[newPosition] = copyCosts[oldIndex];
				population.set(newPosition, copyPopulation.get(oldIndex));
			}
		}
	}
}